@@ -1,5 +1,7 @@
+#ifndef LIBSSH2_OS400QC3_H
+#define LIBSSH2_OS400QC3_H
 /*
- * Copyright (C) 2015 Patrick Monnerat, D+H <patrick.monnerat@dh.com>
+ * Copyright (C) Patrick Monnerat <patrick@monnerat.net>
  * All rights reserved.
  *
  * Redistribution and use in source and binary forms,
@@ -34,10 +36,11 @@
  * NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
  * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY
  * OF SUCH DAMAGE.
+ *
+ * SPDX-License-Identifier: BSD-3-Clause
  */
 
-#ifndef LIBSSH2_OS400QC3_H
-#define LIBSSH2_OS400QC3_H
+#define LIBSSH2_CRYPTO_ENGINE libssh2_os400qc3
 
 #include <stdlib.h>
 #include <string.h>
@@ -160,28 +163,45 @@
 #undef Qc3_MK_Pending
 #define Qc3_MK_Pending      '\xF3'      /* '3' */
 
-
 /* Define which features are supported. */
-#define LIBSSH2_MD5             1
+#ifdef OPENSSL_NO_MD5
+# define LIBSSH2_MD5            0
+#else
+# define LIBSSH2_MD5            1
+#endif
+
 #define LIBSSH2_HMAC_RIPEMD     0
 #define LIBSSH2_HMAC_SHA256     1
 #define LIBSSH2_HMAC_SHA512     1
 
-#define LIBSSH2_AES             1
+#define LIBSSH2_AES_CBC         1
 #define LIBSSH2_AES_CTR         1
+#define LIBSSH2_AES_GCM         0
 #define LIBSSH2_BLOWFISH        0
 #define LIBSSH2_RC4             1
 #define LIBSSH2_CAST            0
 #define LIBSSH2_3DES            1
 
 #define LIBSSH2_RSA             1
+#define LIBSSH2_RSA_SHA1        1
+#define LIBSSH2_RSA_SHA2        1
 #define LIBSSH2_DSA             0
+#define LIBSSH2_ECDSA           0
+#define LIBSSH2_ED25519         0
+
+#include "crypto_config.h"
 
-#define MD5_DIGEST_LENGTH       16
 #define SHA_DIGEST_LENGTH       20
 #define SHA256_DIGEST_LENGTH    32
+#define SHA384_DIGEST_LENGTH    48
 #define SHA512_DIGEST_LENGTH    64
 
+#define EC_MAX_POINT_LEN ((528 * 2 / 8) + 1)
+
+#if LIBSSH2_ECDSA
+#else
+#define _libssh2_ec_key void
+#endif
 
 /*******************************************************************
  *
@@ -210,6 +230,10 @@ typedef struct {        /* Algorithm description. */
     int                     keylen;         /* Key length. */
 }       _libssh2_os400qc3_cipher_t;
 
+typedef struct {        /* Diffie-Hellman context. */
+    char                    token[8];       /* Context token. */
+}       _libssh2_os400qc3_dh_ctx;
+
 /*******************************************************************
  *
  * OS/400 QC3 crypto-library backend: Define global types/codes.
@@ -221,55 +245,50 @@ typedef struct {        /* Algorithm description. */
 
 #define libssh2_sha1_ctx        Qc3_Format_ALGD0100_T
 #define libssh2_sha256_ctx      Qc3_Format_ALGD0100_T
-#define libssh2_md5_ctx         Qc3_Format_ALGD0100_T
+#define libssh2_sha384_ctx      Qc3_Format_ALGD0100_T
+#define libssh2_sha512_ctx      Qc3_Format_ALGD0100_T
 #define libssh2_hmac_ctx        _libssh2_os400qc3_crypto_ctx
 #define _libssh2_cipher_ctx     _libssh2_os400qc3_crypto_ctx
 
-#define libssh2_sha1_init(x)    libssh2_os400qc3_hash_init(x, Qc3_SHA1)
+#define libssh2_sha1_init(x)    _libssh2_os400qc3_hash_init(x, Qc3_SHA1)
 #define libssh2_sha1_update(ctx, data, len)                                 \
-                                libssh2_os400qc3_hash_update(&(ctx), data, len)
+                               _libssh2_os400qc3_hash_update(&(ctx), data, len)
 #define libssh2_sha1_final(ctx, out)                                        \
-                                libssh2_os400qc3_hash_final(&(ctx), out)
-#define libssh2_sha256_init(x)  libssh2_os400qc3_hash_init(x, Qc3_SHA256)
+                                _libssh2_os400qc3_hash_final(&(ctx), out)
+#define libssh2_sha256_init(x)  _libssh2_os400qc3_hash_init(x, Qc3_SHA256)
 #define libssh2_sha256_update(ctx, data, len)                               \
-                                libssh2_os400qc3_hash_update(&(ctx), data, len)
+                               _libssh2_os400qc3_hash_update(&(ctx), data, len)
 #define libssh2_sha256_final(ctx, out)                                      \
-                                libssh2_os400qc3_hash_final(&(ctx), out)
+                                _libssh2_os400qc3_hash_final(&(ctx), out)
 #define libssh2_sha256(message, len, out)                                   \
-                                libssh2_os400qc3_hash(message, len, out,    \
-                                                      Qc3_SHA256)
-#define libssh2_md5_init(x)     libssh2_os400qc3_hash_init(x, Qc3_MD5)
-#define libssh2_md5_update(ctx, data, len)                                  \
-                                libssh2_os400qc3_hash_update(&(ctx), data, len)
-#define libssh2_md5_final(ctx, out)                                         \
-                                libssh2_os400qc3_hash_final(&(ctx), out)
-#define libssh2_hmac_ctx_init(ctx)                                          \
-                                memset((char *) &(ctx), 0,                  \
-                                       sizeof(libssh2_hmac_ctx))
-#define libssh2_hmac_md5_init(ctx, key, keylen)                             \
-                                libssh2_os400qc3_hmac_init(ctx, Qc3_MD5,    \
-                                                           MD5_DIGEST_LENGTH, \
-                                                           key, keylen)
-#define libssh2_hmac_sha1_init(ctx, key, keylen)                            \
-                                libssh2_os400qc3_hmac_init(ctx, Qc3_SHA1,   \
-                                                           SHA_DIGEST_LENGTH, \
-                                                           key, keylen)
-#define libssh2_hmac_sha256_init(ctx, key, keylen)                          \
-                                libssh2_os400qc3_hmac_init(ctx, Qc3_SHA256, \
-                                                         SHA256_DIGEST_LENGTH, \
-                                                         key, keylen)
-#define libssh2_hmac_sha512_init(ctx, key, keylen)                          \
-                                libssh2_os400qc3_hmac_init(ctx, Qc3_SHA512, \
-                                                         SHA512_DIGEST_LENGTH, \
-                                                         key, keylen)
-#define libssh2_hmac_update(ctx, data, datalen)                             \
-                                libssh2_os400qc3_hmac_update(&(ctx),        \
-                                                             data, datalen)
-#define libssh2_hmac_final(ctx, data)                                       \
-                                libssh2_os400qc3_hmac_final(&(ctx), data)
-#define libssh2_hmac_cleanup(ctx)                                           \
-                                _libssh2_os400qc3_crypto_dtor(ctx)
+                                _libssh2_os400qc3_hash(message, len, out,   \
+                                                       Qc3_SHA256)
+#define libssh2_sha384_init(x)  _libssh2_os400qc3_hash_init(x, Qc3_SHA384)
+#define libssh2_sha384_update(ctx, data, len)                               \
+                               _libssh2_os400qc3_hash_update(&(ctx), data, len)
+#define libssh2_sha384_final(ctx, out)                                      \
+                                _libssh2_os400qc3_hash_final(&(ctx), out)
+#define libssh2_sha384(message, len, out)                                   \
+                                _libssh2_os400qc3_hash(message, len, out,   \
+                                                       Qc3_SHA384)
+#define libssh2_sha512_init(x)  _libssh2_os400qc3_hash_init(x, Qc3_SHA512)
+#define libssh2_sha512_update(ctx, data, len)                               \
+                               _libssh2_os400qc3_hash_update(&(ctx), data, len)
+#define libssh2_sha512_final(ctx, out)                                      \
+                                _libssh2_os400qc3_hash_final(&(ctx), out)
+#define libssh2_sha512(message, len, out)                                   \
+                                _libssh2_os400qc3_hash(message, len, out,   \
+                                                       Qc3_SHA512)
 
+#if LIBSSH2_MD5 || LIBSSH2_MD5_PEM
+#define MD5_DIGEST_LENGTH       16
+#define libssh2_md5_ctx         Qc3_Format_ALGD0100_T
+#define libssh2_md5_init(x)     _libssh2_os400qc3_hash_init(x, Qc3_MD5)
+#define libssh2_md5_update(ctx, data, len)                                  \
+                               _libssh2_os400qc3_hash_update(&(ctx), data, len)
+#define libssh2_md5_final(ctx, out)                                         \
+                                _libssh2_os400qc3_hash_final(&(ctx), out)
+#endif
 
 #define _libssh2_bn_ctx         int                 /* Not used. */
 
@@ -277,26 +296,26 @@ typedef struct {        /* Algorithm description. */
 #define _libssh2_bn_ctx_free(bnctx)     ((void) 0)
 
 #define _libssh2_bn_init_from_bin() _libssh2_bn_init()
-#define _libssh2_bn_mod_exp(r, a, p, m, ctx)                                \
-                                _libssh2_os400qc3_bn_mod_exp(r, a, p, m)
 #define _libssh2_bn_bytes(bn)   ((bn)->length)
 
 #define _libssh2_cipher_type(name)  _libssh2_os400qc3_cipher_t name
 #define _libssh2_cipher_aes128 {Qc3_Alg_Block_Cipher, Qc3_AES, 16,          \
                                 Qc3_CBC, 16}
-#define _libssh2_cipher_aes192 {Qc3_Alg_Block_Cipher, Qc3_AES, 24,          \
+#define _libssh2_cipher_aes192 {Qc3_Alg_Block_Cipher, Qc3_AES, 16,          \
                                 Qc3_CBC, 24}
-#define _libssh2_cipher_aes256 {Qc3_Alg_Block_Cipher, Qc3_AES, 32,          \
+#define _libssh2_cipher_aes256 {Qc3_Alg_Block_Cipher, Qc3_AES, 16,          \
                                 Qc3_CBC, 32}
 #define _libssh2_cipher_aes128ctr {Qc3_Alg_Block_Cipher, Qc3_AES, 16,       \
                                    Qc3_CTR, 16}
-#define _libssh2_cipher_aes192ctr {Qc3_Alg_Block_Cipher, Qc3_AES, 24,       \
+#define _libssh2_cipher_aes192ctr {Qc3_Alg_Block_Cipher, Qc3_AES, 16,       \
                                    Qc3_CTR, 24}
-#define _libssh2_cipher_aes256ctr {Qc3_Alg_Block_Cipher, Qc3_AES, 32,       \
+#define _libssh2_cipher_aes256ctr {Qc3_Alg_Block_Cipher, Qc3_AES, 16,       \
                                    Qc3_CTR, 32}
-#define _libssh2_cipher_3des   {Qc3_Alg_Block_Cipher, Qc3_TDES, 0,          \
-                                Qc3_CBC, 24}
-#define _libssh2_cipher_arcfour {Qc3_Alg_Stream_Cipher, Qc3_RC4, 0, 0, 16}
+#define _libssh2_cipher_3des {Qc3_Alg_Block_Cipher, Qc3_TDES, 8,            \
+                              Qc3_CBC, 24}
+/* Nonsense values for chacha20-poly1305 */
+#define _libssh2_cipher_chacha20 {Qc3_Alg_Stream_Cipher, Qc3_RC4, 8, 0, 16}
+#define _libssh2_cipher_arcfour {Qc3_Alg_Stream_Cipher, Qc3_RC4, 8, 0, 16}
 
 #define _libssh2_cipher_dtor(ctx) _libssh2_os400qc3_crypto_dtor(ctx)
 
@@ -306,8 +325,30 @@ typedef struct {        /* Algorithm description. */
 #define libssh2_prepare_iovec(vec, len) memset((char *) (vec), 0,           \
                                                (len) * sizeof(struct iovec))
 #define _libssh2_rsa_sha1_signv(session, sig, siglen, count, vector, ctx)   \
-            _libssh2_os400qc3_rsa_sha1_signv(session, sig, siglen,          \
+            _libssh2_os400qc3_rsa_signv(session, Qc3_SHA1, sig, siglen,     \
                                              count, vector, ctx)
+#define _libssh2_rsa_sha2_256_signv(session, sig, siglen, cnt, vector, ctx) \
+            _libssh2_os400qc3_rsa_signv(session, Qc3_SHA256, sig, siglen,   \
+                                             cnt, vector, ctx)
+#define _libssh2_rsa_sha2_512_signv(session, sig, siglen, cnt, vector, ctx) \
+            _libssh2_os400qc3_rsa_signv(session, Qc3_SHA512, sig, siglen,   \
+                                             cnt, vector, ctx)
+
+/* Default generate and safe prime sizes for diffie-hellman-group-exchange-sha1
+   Qc3 is limited to a maximum 2048-bit modulus/key size. */
+#define LIBSSH2_DH_GEX_MINGROUP     1024
+#define LIBSSH2_DH_GEX_OPTGROUP     1536
+#define LIBSSH2_DH_GEX_MAXGROUP     2048
+
+#define LIBSSH2_DH_MAX_MODULUS_BITS 2048
+
+#define _libssh2_dh_ctx         _libssh2_os400qc3_dh_ctx
+#define libssh2_dh_init(dhctx)  _libssh2_os400qc3_dh_init(dhctx)
+#define libssh2_dh_key_pair(dhctx, public, g, p, group_order, bnctx)        \
+            _libssh2_os400qc3_dh_key_pair(dhctx, public, g, p, group_order)
+#define libssh2_dh_secret(dhctx, secret, f, p, bnctx)                       \
+            _libssh2_os400qc3_dh_secret(dhctx, secret, f, p)
+#define libssh2_dh_dtor(dhctx)  _libssh2_os400qc3_dh_dtor(dhctx)
 
 
 /*******************************************************************
@@ -319,40 +360,38 @@ typedef struct {        /* Algorithm description. */
 extern _libssh2_bn *    _libssh2_bn_init(void);
 extern void     _libssh2_bn_free(_libssh2_bn *bn);
 extern unsigned long    _libssh2_bn_bits(_libssh2_bn *bn);
-extern int      _libssh2_bn_from_bin(_libssh2_bn *bn, int len,
+extern int      _libssh2_bn_from_bin(_libssh2_bn *bn, size_t len,
                                      const unsigned char *v);
 extern int      _libssh2_bn_set_word(_libssh2_bn *bn, unsigned long val);
 extern int      _libssh2_bn_to_bin(_libssh2_bn *bn, unsigned char *val);
-extern void     _libssh2_random(unsigned char *buf, int len);
-extern int      _libssh2_bn_rand(_libssh2_bn *bn, int bits,
-                                int top, int bottom);
-extern int      _libssh2_os400qc3_bn_mod_exp(_libssh2_bn *r, _libssh2_bn *a,
-                                             _libssh2_bn *p, _libssh2_bn *m);
+extern int      _libssh2_random(unsigned char *buf, size_t len);
 extern void     _libssh2_os400qc3_crypto_dtor(_libssh2_os400qc3_crypto_ctx *x);
-extern int      libssh2_os400qc3_hash_init(Qc3_Format_ALGD0100_T *x,
-                                           unsigned int algo);
-extern void     libssh2_os400qc3_hash_update(Qc3_Format_ALGD0100_T *ctx,
-                                             unsigned char *data, int len);
-extern void     libssh2_os400qc3_hash_final(Qc3_Format_ALGD0100_T *ctx,
-                                            unsigned char *out);
-extern int      libssh2_os400qc3_hash(const unsigned char *message,
-                                      unsigned long len, unsigned char *out,
-                                      unsigned int algo);
-extern void     libssh2_os400qc3_hmac_init(_libssh2_os400qc3_crypto_ctx *x,
-                                           int algo, size_t minkeylen,
-                                           void *key, int keylen);
-extern void     libssh2_os400qc3_hmac_update(_libssh2_os400qc3_crypto_ctx *ctx,
-                                             const unsigned char *data,
-                                             int len);
-extern void     libssh2_os400qc3_hmac_final(_libssh2_os400qc3_crypto_ctx *ctx,
-                                            unsigned char *out);
-extern int      _libssh2_os400qc3_rsa_sha1_signv(LIBSSH2_SESSION *session,
-                                                 unsigned char **signature,
-                                                 size_t *signature_len,
-                                                 int veccount,
-                                                 const struct iovec vector[],
-                                                 libssh2_rsa_ctx *ctx);
+extern int      _libssh2_os400qc3_hash_init(Qc3_Format_ALGD0100_T *x,
+                                            unsigned int algo);
+extern int      _libssh2_os400qc3_hash_update(Qc3_Format_ALGD0100_T *ctx,
+                                              const unsigned char *data,
+                                              int len);
+extern int      _libssh2_os400qc3_hash_final(Qc3_Format_ALGD0100_T *ctx,
+                                             unsigned char *out);
+extern int      _libssh2_os400qc3_hash(const unsigned char *message,
+                                       unsigned long len, unsigned char *out,
+                                       unsigned int algo);
+extern int      _libssh2_os400qc3_rsa_signv(LIBSSH2_SESSION *session, int algo,
+                                            unsigned char **signature,
+                                            size_t *signature_len,
+                                            int veccount,
+                                            const struct iovec vector[],
+                                            libssh2_rsa_ctx *ctx);
+extern void     _libssh2_os400qc3_dh_init(_libssh2_dh_ctx *dhctx);
+extern int      _libssh2_os400qc3_dh_key_pair(_libssh2_dh_ctx *dhctx,
+                                              _libssh2_bn *public,
+                                              _libssh2_bn *g,
+                                              _libssh2_bn *p, int group_order);
+extern int      _libssh2_os400qc3_dh_secret(_libssh2_dh_ctx *dhctx,
+                                            _libssh2_bn *secret,
+                                            _libssh2_bn *f, _libssh2_bn *p);
+extern void     _libssh2_os400qc3_dh_dtor(_libssh2_dh_ctx *dhctx);
 
-#endif
+#endif /* LIBSSH2_OS400QC3_H */
 
 /* vim: set expandtab ts=4 sw=4: */
